#include "curve.h"

// Draws a cubic curve defined by parametric equations x(t), y(t), z(t).
// x(t) = at^3 + bt^2 + ct + d.  y(t) and z(t) have similar definitions with different
// coefficients.  The coefficients are in the x, y, and z arrays.  The curve will be a
// line strip made up of n lines.
void cubicCurve(GLfloat *x, GLfloat *y, GLfloat *z, GLint n)
{	GLfloat t = 0, t2 = 0, t3 = 0;
	GLfloat increment = 1.0 / n;
	
	glBegin(GL_LINE_STRIP);
		for(int i = 0; i <= n; i++)
		{	glVertex3f(	x[0] * t3 + x[1] * t2 + x[2] * t + x[3],
						y[0] * t3 + y[1] * t2 + y[2] * t + y[3],
						z[0] * t3 + z[1] * t2 + z[2] * t + z[3]);
			t += increment;
			t2 = t * t;
			t3 = t * t2;
		}
	glEnd();
}

// Returns the distance between (x1, y1) and (x2, y2).
float dist(int x1, int y1, int x2, int y2)
{	return sqrt((x1-x2)*(x1-x2)+(y1-y2)*(y1-y2));
}

// Returns the distance between the two coordinates passed.  (Each has an X, Y, and Z.)
GLfloat dist(GLfloat * one, GLfloat * two)
{	GLfloat x = one[0] - two[0];
	GLfloat y = one[1] - two[1];	
	GLfloat z = one[2] - two[2];
	return sqrt(x*x + y*y + z*z);
}


// Now the curve class functions:

// Constructor with basic defaults.  Children should take care of everything, though
curve::curve(): showPoints(true), detail(50), number(0), dirty(true), points(NULL)
{	M = new basis[1];				// One basis matrix, all 0s.
	
	for(int i = 0; i < 3; i++)
		coeffs[i] = NULL;
}

// *************** Will this work if they weren't allocated in this way? ********************
// Destructor:  Make sure no dynamic memory is associated with our pointers
curve::~curve()
{	delete [] M;
	delete [] points;
	for(int i = 0; i < 3; i++)
		delete [] coeffs[i];
}

// Returns the X, Y, or Z value (as specified by which) of control point #index.
// If the requested point does not exist, an error is printed and 0 is returned.
GLfloat curve::get(unsigned int index, coordEnum which) const
{	if(index >= number)
	{	cerr << "Illegal value passed to curve::get().\n";
		return 0;
	}
	return points[index][which];
}

// Copies the X, Y, and Z values (in that order) of control point #index into the
// specified array.  If the requested point does not exist, an error is printed
// and no values are copied.
void curve::get(unsigned int index, GLfloat *array) const
{	if(index >= number)
	{	cerr << "Illegal value passed to curve::get().\n";
		return;
	}
	for(int i = 0; i < 3; i++)
		array[i] = points[index][i];
}

// Returns number.
int curve::howManyPoints() const
{	return number;
}

// Returns the showPoints data.  This is a boolean that is true if the control points are to
// be drawn with the curve and false if not.
bool curve::displayingPoints() const
{	return showPoints;
}

// Returns the detail of the curve.  This value gives the accuracy of the approximation
// when this curve is drawn.  Generally, this is the number of line segments used.
int curve::getDetail() const
{	return detail;
}

// Set the X, Y, or Z value (as specified by which) of control point # index to the specified value.
// If index does not refer to a valid control point, an error is printed and nothing is changed.
void curve::set(unsigned int index, coordEnum which, GLfloat value)
{	if(index >= number)
	{	cerr << "Error.  Illegal value passed to curve::set().\n";
		return;
	}
	points[index][which] = value;
	
	dirty = true;							// The coefficients need to be recalculated
}

// Changes the values of control point #index to those passed in the array.  As you should be
// able to guess by now, an error will be printed and nothing changed if index is not valid.
void curve::set(unsigned int index, GLfloat *array)
{	if(index >= number)
	{	cerr << "Error.  Illegal value passed to curve::set().\n";
		return;
	}
	for(int i = 0; i < 3; i++)
		points[index][i] = array[i];
	dirty = true;							// The coefficients need to be recalculated
}

// This function sets all the control points to the values in the array.  The total number of control
// points remains unchanged.  No error checking is done to make sure that the user has
// provided enough values -- Everything accessed is considered to be a valid GLfloat.
void curve::setAll(GLfloat * array)
{	for(int i = 0; i <number; i++)
		for(int j = 0; j < 3; j++)
			points[i][j] = array[i*3 + j];
	dirty = true;							// The coefficients need to be recalculated
}

// This function changes the number of control points to newnum and gives them the values
// passed in the array.  It will also, if necessary, reallocate memory for coeffs.  (The assumption
// is that every segment is made up of four points with a single one shared between segments.)
void curve::setAll(unsigned int newnum, GLfloat * array)
{	if(newnum != number)
	{	if(number > 0)						// If there are already points, we first need to
		{	delete [] points;					// deallocate the memory of them.  Don't worry
			for(int i = 0; i < 3; i++)			// about saving it...everything will change.
				delete [] coeffs[i];	
		}
		if(newnum > 0)						// If there are going to be any points, we need
		{	points = new coordinate[newnum];	// to allocate memory for them.
			if(newnum >= 3)				// And if there are enough points for a curve
				for(int i = 0; i < 3; i++)		// segment, allocate space for them, too.
					coeffs[i] = new coefficient[newnum / 3];
		}
		number = newnum;
	}
	
	// Whatever happened above, we know that points and number are now correct.
	for(int i = 0; i <number; i++)				// Copy values of control points
		for(int j = 0; j < 3; j++)
			points[i][j] = array[i*3 + j];
	dirty = true;							// The coefficients need to be recalculated
}

// This function adds a control point to the curve, situated on the origin.  If we now have enough
// control points to make a new segment, do so.  (Points 0..3 make the first, 3..6 make the second,
// etc.  So it is sufficient to see if the number of points, counting from 0, is divisible by 3.  We subtract
// 1 from the number variable, since it normally counts from 1.)
void curve::addPoint()
{	if(number > 0)							// Save the points that already exist.
	{	coordinate * temp = new coordinate[number+1];
		for(int i = 0; i < number; i++)
			for(int j = 0; j < 3; j++)
				temp[i][j] = points[i][j];
		
		delete [] points;
		points = temp;
	}
	else									// Otherwise, we just need room for one new
		points = new coordinate[1];			// point.
	
	
	number++;							// Increase our count of control points.
	for(int i = 0; i < 3; i++)					// The new control point defaults to the origin
		points[number-1][i] = 0;
	
	
	if(number%3 == 0)						// We now have enough for a new segment
	{	if(number > 3)						// First, we need to deallocate the old coefficient
			for(int i = 0; i < 3; i++)			// info if it exists (ie, if this isn't the first segment.)
				delete [] coeffs[i];			// Then we allocate new memory for them and
		for(int i = 0; i < 3; i++)				// calculate their values.  (Seperate loops are
			coeffs[i] = new coefficient[number/3];// used so that as much memory as possible
										// is freed before we allocate more.)
		dirty = true;						// The coefficients need to be recalculated
	}
}

// This function adds a control point to the curve, with a location given by the array.  If we now have
// enough control points to make a new segment, do so.  (Points 0..3 define the first, 3..6 define the
// second, etc.  So it is sufficient to see if the number of points, counting from 0, is divisible by 3.  We
// subtract 1 from the number variable, since it normally counts from 1.)
void curve::addPoint(GLfloat * array)
{	if(number > 0)							// Save the points that already exist.
	{	coordinate * temp = new coordinate[number+1];
		for(int i = 0; i < number; i++)
			for(int j = 0; j < 3; j++)
				temp[i][j] = points[i][j];
		
		delete [] points;
		points = temp;
	}
	else									// Otherwise, we just need room for one new
		points = new coordinate[1];			// point.
	
	
	number++;							// Increase our count of control points.
	for(int i = 0; i < 3; i++)					// The new control point defaults to the origin
		points[number-1][i] = array[i];
	
	
	if(number%3 == 0)						// We now have enough for a new segment
	{	if(number > 3)						// First, we need to deallocate the old coefficient
			for(int i = 0; i < 3; i++)			// info if it exists (ie, if this isn't the first segment.)
				delete [] coeffs[i];			// Then we allocate new memory for them
		for(int i = 0; i < 3; i++)				// and calculate their values.
			coeffs[i] = new coefficient[number/3];
		dirty = true;						// The coefficients need to be recalculated
	}
}

// Set whether or not the control points of the curve should be displayed when the rest of the
// curve is.
void curve::displayPoints(bool arg)
{	showPoints = arg;
}

// Sets the detail with which the curve is drawn.  The exact meaning of this may very between
// different representations of curves, but it will often be the number of line segments used per
// curve segment.  This function makes the assumption that only values greater than 0 are
// valid detail levels.  If a child class needs special error checking or functionality, they must
// overload this.  For consistency with child classes that might accept values < 0, the parameter
// this function takes is an int rather than an unsigned one.
// No error is given if detail <= 0, but nothing is changed either.
void curve::setDetail(int arg)
{	if(arg > 0)
		detail = arg;
}

// Calculate the coefficients of the cubic equations of the curve.  Each segment will have a set of
// 4 coefficients that correspond to 1, t, t^2, and t^3.  We calculate this as M*G.  (The geometry
// vector is found by looking "down" the list of points.  For example, the four x values of the control
// points give you the x geometry vector.)
// At the end of this function, we mark the curve as "clean".  That is, the coefficients have been
// updated to correspond with the curve's properties.  This function will not need to be called again
// until something is changed (setting the dirty flag) and draw() or getAtTimeT() is needs the
// updated properties.
void curve::calcCoeffs()
{	coefficient G;								// Put geometry vector in this
	for(int coord = 0; coord < 3; coord++)			// loop for each of X, Y, and Z
		for(int seg = 0; seg < (number-1)/3; seg++)	// and for each individual segment
		{	for(int i = 0; i < 4; i++)
				G[i] = points[seg*3+i][coord];		// Make geometry vector
			coeffs[coord][seg] = M[0] * G;			// And copy it
		}
/*			for(int i = 0; i < 4; i++)			// Calculate each of the four coefficients
				coeffs[coord][seg][i] = points[seg*3][coord] * M[0][0][i] + 
								  points[seg*3+1][coord] * M[0][1][i] + 
								  points[seg*3+2][coord] * M[0][2][i] + 
								  points[seg*3+3][coord] * M[0][3][i];*/
	dirty = false;							// Curve is now "clean".
}