#include "hermite.h"

// Default constructor.  It creates a "curve" with 0 control points.  The only real duties of this function
// are to give everything default values, allocate a single basis matrix, and give it its values.
hermite::hermite(): cont(G0)
{	GLfloat mat[16]  =  {   2, 1, 1,-2,					// Hermite basis matrix
					-3,-2,-1, 3,
					 0, 1, 0, 0,
					 1, 0, 0, 0 };
	M = new basis[1];							// Allocate basis matrix
	M[0] = mat;								// and give it values.
	
	showPoints = true;
	detail = 50;
	number = 0;
	dirty = true;
	points = NULL;
	for(int i = 0; i < 3; i++)
		coeffs[i] = NULL;
}

// This constructor sets up the appropriate number of points, but they are all currently
// on the origin.  The curve won't be very interesting until they are changed.
// The hermite class is unusual in that it only works with exact numbers of points needed to
// make a segment, and extra points on the end are not allowed.  For this reason, num may be
// reduced to a lower value.
hermite::hermite(unsigned int num) : cont(G0)
{	GLfloat mat[16]  =  {   2, 1, 1,-2,					// Hermite basis matrix
					-3,-2,-1, 3,
					 0, 1, 0, 0,
					 1, 0, 0, 0 };
	if(num > 0)
		num -= (num-1)%3;						// Get rid of extra points.
	number = num;
	
	M = new basis[1];							// Allocate basis matrix
	M[0] = mat;								// and give it values.
	
	if(num > 0)								// Now allocate space to store the
	{	points = new coordinate[number];			// control points and coefficients
		if(num > 3)							// (if necessary).
			for(int i = 0; i < 3; i++)
				coeffs[i] = new coefficient[(number-1)/3];
		else coeffs[0] = coeffs[1] = coeffs[2] = NULL;
	}
	else points = NULL;
	
	for(int i = 0; i < number; i++)					// Initialize all points to origin
		for(int j = 0; j < 3; j++)
			points[i][j] = 0;
	
	showPoints = true;
	detail = 50;
	dirty = true;
}

// This constructor sets up a curve with the specified number of points whose values are
// provided by the array.
// The hermite class is unusual in that it only works with exact numbers of points needed to
// make a segment, and extra points on the end are not allowed.  For this reason, num may be
// reduced to a lower value.
hermite::hermite(unsigned int num, GLfloat * array) : cont (G0)
{	GLfloat mat[16]  =  {   2, 1, 1,-2,					// Hermite basis matrix
					-3,-2,-1, 3,
					 0, 1, 0, 0,
					 1, 0, 0, 0 };
	if(num > 0)
		num -= (num-1)%3;						// Get rid of extra points.
	number = num;
	
	M = new basis[1];							// Allocate basis matrix
	M[0] = mat;								// and give it values.
	
	if(num > 0)								// Now allocate space to store the
	{	points = new coordinate[number];			// control points and coefficients
		if(num > 3)							// (if necessary).
			for(int i = 0; i < 3; i++)
				coeffs[i] = new coefficient[(number-1)/3];
		else coeffs[0] = coeffs[1] = coeffs[2] = NULL;
	}
	else points = NULL;
	
	for(int i = 0; i < number; i++)					// Initialize all points to values passed
		for(int j = 0; j < 3; j++)					// in the array.
			points[i][j] = array[i*3+j];
	
	showPoints = true;
	detail = 50;
	dirty = true;
}

// Returns the continuity of the curve
Continuity hermite::getContinuity() const
{	return cont;
}

// This returns the number of segments in the complete cuve.  The first four points define one
// segment, and because of the shared control point, every additional three create one more.
// So just subtract 1 from the number of points and divide it by 3.
int hermite::howManySegments() const
{	return (number - 1) / 3;
}

// The hermite class must overload the functions that set values to make sure that continuity is
// retained.  Also, when an endpoint is moved, its associated tangent "control points" are moved
// as well so that the tangent vectors are unchanged.
// With the exception of those considerations, the logic behind these set*() functions is the same
// as that of the parent class.
void hermite::set(unsigned int index, coordEnum which, GLfloat value)
{	if(index >= number)
	{	cerr << "Error.  Illegal value passed to hermite::set().\n";
		return;
	}
	
	dirty = true;							// Mark the curve as changed.
	
	if(index%3 == 0)						// If we are moving an endpoint, keep the
	{	value -= points[index][which];			// tangent vectors constant by turning it into
		points[index][which] += value;			// an offset and adding it to all affected points.
		if(index > 0)						// Check to make sure that this isn't the very
			points[index-1][which] += value;	// first or very last point, because we don't
		if (index < number - 1)				// want to change values beyond the array.
			points[index+1][which] += value;
		return;							// That's all...continuity has also been assured.
	}
	
	// Othewise, we need to change a tangent vector and (possibly) ensure continuity:
	points[index][which] = value;
	
	if(cont == G0 || index < 2)					// In this case, continuity is not a concern.
		return;
	
	if(index % 3 == 1)						// This is an R0...fix the R3 in the previous
	{	if(index > 1)						// segment (if there is a previous one.)
			enforceCont(index-1, -1);
	}
	else									// If none of the above were true, this must be
		if(index < number - 3)				// an R3...fix the tangent to the "right" IF there are
			enforceCont(index+1, 1);			// at least two more points to the right...
}

// Changes the values of control point #index to those passed in the array.  An error will be
// printed and nothing changed if index is not valid.
void hermite::set(unsigned int index, GLfloat *array)
{	if(index >= number)
	{	cerr << "Error.  Illegal value passed to hermite::set().\n";
		return;
	}
	
	dirty = true;							// Mark the curve as changed.
	
	if(index%3 == 0)						// If we are moving an endpoint, keep the
	{	GLfloat offset[3];					// tangent vectors constant by calculating the
		for(int i = 0; i < 3; i++)				// distance it is being moved, and then move
		{	offset[i] = array[i] - points[index][i];	// the associated tangent vectors by the
			points[index][i] += offset[i];		// same amount.  But remember to check
			if(index > 0)					// if this is the very first or very last point, 
				points[index-1][i] += offset[i];	// because we don't want to change values
			if (index < number - 1)			// beyond the array.
				points[index+1][i] += offset[i];
		}
		return;							// That's all...continuity has also been assured.
	}
	
	// Othewise, we need to change a tangent vector and (possibly) ensure continuity:
	for(int i = 0; i < 3; i++)
		points[index][i] = array[i];
	
	if(cont == G0 || index < 2)					// In this case, continuity is not a concern.
		return;
	
	if(index % 3 == 1)						// This is an R0...fix the R3 in the previous
	{	if(index > 1)						// segment (if there is a previous one.)
			enforceCont(index-1, -1);
	}
	else									// If none of the above were true, this must be
		if(index < number - 3)				// an R3...fix the tangent to the "right" IF there are
			enforceCont(index+1, 1);			// at least two more points to the right...
}

// This function sets all the control points to the values in the array.  The total number of control
// points remains unchanged.  No error checking is done to make sure that the user has
// provided enough values -- Everything accessed is considered to be a valid GLfloat.
// If the curve is C1 or G1 continuous, then the continuity is preserved by changing the values to the
// "left" of the internal join points.  Yes, this ignores some of the user's input, but continuity is
// necessarily restrictive.
void hermite::setAll(GLfloat * array)
{	for(int i = 0; i <number; i++)
		for(int j = 0; j < 3; j++)
			points[i][j] = array[i*3 + j];
	if(cont != G0)							// Do we need to make the curve continuous?
		for(int i = 3; i < number - 1; i++)			// For each of the internal join points...
			enforceCont(i, -1);				// alter the points to be continuous.
	dirty = true;							// The coefficients need to be recalculated
}

// This function changes the number of control points to newnum and gives them the values
// passed in the array.  It will also, if necessary, reallocate memory for coeffs.  (Remember
// that every segment is made up of four points with a single one shared between segments.)
// If the curve is C1 or G1 continuous, then the continuity is preserved by changing the values to the
// "left" of the internal join points.  This ignores some of the user's input, but continuity is
// necessarily restrictive.
// In addition, this program thinks in terms of segments.  So if the user has requested a few extra
// points that don't make an exact segment, then they are ignored.
void hermite::setAll(unsigned int newnum, GLfloat * array)
{	if(newnum > 0)
		newnum -= (newnum-1)%3;			// Ignore extra points that don't define a segment
	if(newnum != number)
	{	if(number > 0)						// If there are already points, we first need to
		{	delete [] points;					// deallocate the memory for them.  Don't worry
			for(int i = 0; i < 3; i++)			// about saving it...everything will change.
				delete [] coeffs[i];	
		}
		if(newnum > 0)						// If there are going to be any points, we need
		{	points = new coordinate[newnum];	// to allocate memory for them.
			if(newnum >=3)				// And if there are enough points for a curve
				for(int i = 0; i < 3; i++)		// segment, allocate space for them, too.
					coeffs[i] = new coefficient[(newnum-1) / 3];
		}
		number = newnum;
	}
	
	// Whatever happened above, we know that points and number are now correct.
	for(int i = 0; i <number; i++)				// Copy values of control points
		for(int j = 0; j < 3; j++)
			points[i][j] = array[i*3 + j];
	if(cont != G0)							// Do we need to make the curve continuous?
		for(int i = 3; i < number - 1; i++)			// For each of the internal join points...
			enforceCont(i, -1);				// alter the points to be continuous.
	dirty = true;							// The coefficients need to be recalculated
}

// This function adds a new segment to the curve.  This is a bit different than the other versions of the
// curve classes, since they only add one control point at a time, but since the Hermite tangent vectors
// aren't control points, this is being treated differently.
// An endpoint is placed at the origin (since no other location is provided), and the tangent vectors are
// just (10, 10, 0) if there is no continuity.  The user can move these wherever is desired.
// If this is the first point, then a start point is put at the origin and no tangent vectors are created.
void hermite::addPoint()
{	coordinate * temp;						// Used for copying points
	if(number == 0)						// Special case:  Adding the first point
	{	points = new coordinate[1];			// Allocate space for start point.
		for(int i = 0; i < 3; i++)				// (Before this, points and coeffs were both
			points[0][i] = 0;					// NULL.  coeffs should remain this way.  This
		number = 1;						// simplifies things.)
		return;
	}
	
	// Now the default case:
	temp = new coordinate[number+3];			// Two new tangents and an endpoint
	for(int i = 0; i < number; i++)
		for(int j = 0; j < 3; j++)
			temp[i][j] = points[i][j];			// Copy all original points to new array.
	for(int i = 0; i < 2; i++)					// Now make values for the two new tangent
	{	temp[number][i] = temp[number-1][i]+10;	// vectors and the endpoint.
		temp[number+1][i] = 10;
		temp[number+2][i] = 0;
	}
	temp[number][2] = temp[number-1][2];
	temp[number+1][2] = temp[number+2][2] = 0;
	
	delete [] points;							// Now reorganize the memory allocation,
	points = temp;							// redirect the pointers, update the count of
	if(number >= 4)						// points, etc.
		for(int i = 0; i < 3; i++)
			delete [] coeffs[i];
	number += 3;
	for(int i = 0; i < 3; i++)
		coeffs[i] = new coefficient[(number-1)/3];

	if(cont != G0 && number >=4)				// If continuity between segments is a concern...
		enforceCont(number-1, 1);			// then move the new tangent vector accordingly.
	
	dirty = true;							// Set the dirty bit.
}

// This function adds a new segment to the curve.  This is a bit different than the other versions of the
// curve classes, since they only add one control point at a time, but since the Hermite tangent vectors
// aren't control points, this is being treated differently.
// The position of the new endpoint is specified by the array argument, and the tangent vectors are
// just (10, 10, 0) unless there are continuity concerns.  The user can easily change the tangents after
// this finishes.
// If this is the first point, then a start point is put at the origin and no tangents are added.
void hermite::addPoint(GLfloat * array)
{	coordinate * temp;						// Used for copying points
	if(number == 0)						// Special case:  Adding the first point
	{	points = new coordinate[1];			// Allocate space for start point.
		for(int i = 0; i < 3; i++)				// (Before this, points and coeffs were both
			points[0][i] = array[i];				// NULL.  coeffs should remain this way.  This
		number = 1;						// simplifies things.)
		return;
	}
	
	// Now the default case:
	temp = new coordinate[number+3];			// Two new tangents and an endpoint
	for(int i = 0; i < number; i++)
		for(int j = 0; j < 3; j++)
			temp[i][j] = points[i][j];			// Copy all original points to new array.
	for(int i = 0; i < 2; i++)					// Now make values for the two new tangent
	{	temp[number][i] = temp[number-1][i]+10;	// vectors and the endpoint.
		temp[number+1][i] = 10+array[i];
		temp[number+2][i] = array[i];
	}
	temp[number][2] = temp[number-1][2];
	temp[number+1][2] = temp[number+2][2] = array[2];
	
	delete [] points;							// Now reorganize the memory allocation,
	points = temp;							// redirect the pointers, update the count of
	if(number >= 4)						// points, etc.
		for(int i = 0; i < 3; i++)
			delete [] coeffs[i];
	number += 3;
	for(int i = 0; i < 3; i++)
		coeffs[i] = new coefficient[(number-1)/3];

	if(cont != G0 && number >=4)				// If continuity between segments is a concern...
		enforceCont(number-1, 1);			// then move the new tangent vector accordingly.
	
	dirty = true;							// Set the dirty bit.
}

// This function changes the curve's continuity of its join points to the requested degree.
void hermite::setContinuity(Continuity arg)
{	if(arg == cont)						// If nothing's being changed, leave the function
		return;						// without going through all the trouble below.
	switch(arg)
	{	case G0:						// If G0, we don't need to change anything.  So the
			cont = G0;					// points aren't moved and the dirty flag isn't set.
			break;
		case G1:						// For G1 and C1 continuity, we move the points
			cont = G1;					// to the "right" of the join points so that we have the
			for(int i = 3; i < number-2; i+=3)// needed continuity.  These for() loops iterate once
				enforceCont(i, -1);		// for each internal joint (3, 6, 9, . . . as long as there
			dirty = true;				// are at least 3 more control points to the right.  This
			break;					// end condition makes sure that there is a segment
		case C1:						// on both sides of the joint.)  Then it is simply
			cont = C1;					// necessary to call the enforceCont() function, which
			for(int i = 3; i < number-2; i+=3)// moves the control point for us.
				enforceCont(i, -1);
			dirty = true;
			break;
	}
}

// Calculate the value of the equation for a specific value of t, and return the value of the desired
// coordinate (either X, Y, or Z).  If t < 0, then it shall be assumed that the first segment stretches
// to -.  If t > (number-1)/3 then it shall be assumed that the last segment goes to +.  If there
// is no valid segment, 0 is returned.
GLfloat hermite::getAtTimeT(coordEnum which, GLfloat t)
{	int segment = t;					// integer value of t is which segment we use.
	GLfloat localT = t - segment;			// Value of t within segment, normalized to [0..1]
	GLfloat t2, t3;						// localT^2 and localT^3, with easier names.
	
	if (number < 4)						// If there aren't any segments yet, return 0
		return 0;
	
	if(segment < 0)						// Can't be before the 0th segment
	{	segment = 0;
		localT = t;
	}
	else if(segment > (number-1)/3)		// Can't be beyond last segment
	{	segment = (number-1)/3;
		localT = t - segment;
	}
	
	if(dirty)							// Do we need to recalculate the coefficients?
		calcCoeffs();
	
	t2 = localT*localT;
	t3 = t2 * localT;
	
	return (coeffs[which][segment][0]*t3 + coeffs[which][segment][1]*t2 + 
		    coeffs[which][segment][2]*localT + coeffs[which][segment][3]);
}

// Calculate the value of the equation for a specific value of t.  The answer is put in the array that
// has been passed to the function.  If t < 0, then it shall be assumed that the first segment
// stretches to -.  If t > (number-1)/3 then it shall be assumed that the last segment goes to +.
// If there is no valid segment, (0, 0, 0) is returned.
void hermite::getAtTimeT(GLfloat t, GLfloat * array)
{	int segment = t;					// integer value of t is which segment we use.
	GLfloat localT = t - segment;			// Value of t within segment, normalized to [0..1]
	GLfloat t2, t3;						// localT^2 and localT^3, with easier names.
	
	if (number < 4)						// If there aren't any segments yet, return (0, 0, 0)
		for(int i = 0; i < 3; i++)
			array[i] = 0;
	
	if(segment < 0)						// Can't be before the 0th segment
	{	segment = 0;
		localT = t;
	}
	else if(segment >= (number-1)/3)		// Can't be beyond last segment
	{	segment = (number-1)/3 - 1;
		localT = t - segment;
	}
	
	if(dirty)							// Do we need to recalculate the coefficients?
		calcCoeffs();
	
	t2 = localT*localT;
	t3 = t2 * localT;
	
	for(int i = 0; i < 3; i++)				// Calculate for the given t
		array[i] = coeffs[i][segment][0]*t3 + coeffs[i][segment][1]*t2 + 
			        coeffs[i][segment][2]*localT + coeffs[i][segment][3];
}

// Draw the curve and possibly its tangent vectors, which are arrows starting at the endpoint.
void hermite::draw()
{	if(dirty)							// If the values of the curve have been changed,
		calcCoeffs();					// then calculate the new coefficients.
	
	if(showPoints)						// If we are showing the control points,
	{	glColor3f(0.5, 0.5, 0.7);			// Set a bluish-gray color for drawing arrows.
		int segments = (number-1)/3;		// Number of segments in the curve.
		glBegin(GL_LINES);				// Tell OpenGL to start making line segments
		for(int i = 0; i+3 < number; i+=3)	// Draw two arrows for each segment:
		{	drawArrow(points[i], points[i+1]);
			drawArrow(points[i+3], points[i+2]);
		}
		glEnd();
	}
	
	glColor3f(1.0, 1.0, 1.0);				// Set the color to white
	for(int i = 0; i < (number-1)/3; i++)		// Now draw the curve segments themselves.
		cubicCurve(&coeffs[0][i][0], &coeffs[1][i][0], &coeffs[2][i][0], detail);
}

// This function is used to keep the curve continuous.  When a change is made that could break
// the continuity, this function should be called.  The first parameter indexes the start/endpoint of
// a segment around which the continuity is threatened.  To preserve continuity, one of the two
// points "next" to that control point must be moved.  The second parameter, which must be
// either 1 or -1 (because this is a protected function, the assumption is that the arguments are
// valid) tells which one this is.  Point #(index+change) will be changed so that it meets the
// requirements set by point #index and point #(index-change).
void hermite::enforceCont(unsigned int index, int change)
{	GLfloat dist1, dist2;					// These will hold distances between points
	switch(cont)						// How we do this depends on the degree of
	{	case G0:						// continuity we are requiring.
			break;					// G0 (or C0) requires no effort.
		// For G1 continuity, the three points determining the shape of the curve at the joint
		// must be colinear with the middle point being on one side of the tangent endpoints.
		case G1:	
			dist1 = dist(points[index], points[index-change]);	// Dist to point that won't move
			dist2 = dist(points[index], points[index+change]);	// Preserve this distance
			for(int i = 0; i < 3; i++)						// while moving the point
				points[index+change][i] =
					points[index][i]+(points[index-change][i]-points[index][i])*(dist2/dist1);
			break;
		// C1 continuity is easy.  Both points must have the same location.
		case C1:
			for(int i = 0; i < 3; i++)
				points[index+change][i] = points[index-change][i];
			break;
	}
	dirty = true;
}

// Calculate the coefficients of the cubic equations of the curve.  Each segment will have a set of
// 4 coefficients that correspond to 1, t, t^2, and t^3.  We calculate this as M*G.  (The geometry
// vector is found by looking "down" the list of points.  For example, the four x values of the control
// points give you the x geometry vector.)
// At the end of this function, we mark the curve as "clean".  That is, the coefficients have been
// updated to correspond with the curve's properties.  This function will not need to be called again
// until something is changed (setting the dirty flag) and draw() or getAtTimeT() is needs the
// updated properties.
// The difference between this function and the parent calcCoeffs() is that this one accounts for
// the offset that has been added to all the tangent vectors.
void hermite::calcCoeffs()
{	coefficient G;								// Put geometry vector in this
	for(int coord = 0; coord < 3; coord++)			// loop for each of X, Y, and Z
		for(int seg = 0; seg < (number-1)/3; seg++)	// and for each individual segment
		{	for(int i = 0; i < 4; i++)
				G[i] = points[seg*3+i][coord];		// Make geometry vector
			G[1] -= G[0];						// Turn the pseudo-control-points into
			G[2] -= G[3];						// real tangent vectors.
			coeffs[coord][seg] = M[0] * G;			// And copy it
		}
	dirty = false;							// Curve is now "clean".
}

// This (non-class) function calculates the vertices necessary to draw an arrow from point A to point B.
// Just the glVertex calls are made here, with the assumption that glBegin(GL_LINES) was called
// beforehand.
// The little lines that make up the arrowhead have a length of ARROW_HEAD_LENGTH, and they are
// oriented so that they are seen in the x-y plane.
void drawArrow(GLfloat * A, GLfloat * B)
{	GLfloat vector[3];					// A vector pointing from the end to the start
	GLfloat magnitude=0;				// length of vector.  Must be normalized.
	for(int i = 0; i < 3; i++)				// To make arrowhead, we need a vector pointing from
	{	vector[i] = A[i] - B[i];				// B to A.
		magnitude += vector[i]*vector[i];	// Also calculate the magnitude of the vector.
	}
	magnitude = sqrt (magnitude);			// Finish calculating magnitude of vector.
	for(int i = 0; i < 3; i++)
		vector[i] *= ARROW_HEAD_LENGTH/magnitude;	// Now normalize length of vector.
	
	glVertex3fv(A);						// Start out by drawing the line connecting the points
	glVertex3fv(B);
glEnd();	
	glPushMatrix();						// Save current matrix...we're gonna do fun stuff.
		glTranslatef(B[0], B[1], B[2]);		// Translate coordinate system to put B at origin.
		glRotatef(45.0, 0, 0, 1.0);			// Rotate around point B.
glBegin(GL_LINES);
		glVertex3fv(vector);				// And draw a line from B to the end of the vector.
		glVertex3f(0, 0, 0);
glEnd();
	glPopMatrix();
	glPushMatrix();
		glTranslatef(B[0], B[1], B[2]);		// Translate coordinate system to put B at origin.
		glRotatef(-45, 0, 0, 1.0);			// Now rotate back around to draw the other line.
glBegin(GL_LINES);
		glVertex3fv(vector);
		glVertex3f(0, 0, 0);
glEnd();
	glPopMatrix();						// Restore Modelview matrix to its former state.
glBegin(GL_LINES);
}