#include "matrix.h"

// Generic constructor.  Creates a 0x0 matrix
matrix::matrix(): rows(0), columns(0)
{	values = NULL;	}

// Constructor:  # of rows and columns is provided, but all cells are initialized to 0
matrix::matrix(unsigned int r, unsigned int c): rows(r), columns(c)
{	int total = r*c;							// Total number of elements in matrix
	if (total > 0)
	{	values = new GLfloat[total];
		for(int i = 0; i < total; i++)
			values[i] = 0;
	}
	else values = NULL;
}

// Constructor:  User has kindly provided dimensions and initial values for our array.
matrix::matrix(unsigned int r, unsigned int c, GLfloat * array): rows(r), columns(c)
{	int total = r*c;							// Total number of elements in matrix
	if(total > 0)
	{	values = new GLfloat[total];
		for(int i = 0; i < total; i++)
			values[i] = array[i];
	}
	else values = NULL;
}

// Copy Constructor:
matrix::matrix(const matrix & arg): rows(arg.rows), columns(arg.columns)
{	int total = rows*columns;					// Total number of elements in matrix
	if (total > 0)
	{	values = new GLfloat[total];
		for(int i = 0; i < total; i++)
			values[i] = arg.values[i];
	}
	else values = NULL;
}

// Destructor:  Free the memory allocated for values
matrix::~matrix()
{	delete [] values;
}

// Constructors for Child Classes:
coefficient::coefficient()
{	rows = 4;
	columns = 1;
	values = new GLfloat[4];
	for(int i = 0; i < 4; i++)
		values[i] = 0;
}

coefficient::coefficient(const GLfloat * array)
{	rows = 4;
	columns = 1;
	values = new GLfloat[4];
	for(int i = 0; i < 4; i++)
		values[i] = array[i];
}

coefficient::coefficient(const matrix & arg)
{	int total = arg.getRows() * arg.getColumns();	// Total number of elements
	rows = arg.getRows();
	columns = arg.getColumns();
	values = new GLfloat[total];
	for(int i = 0; i < total; i++)
		values[i] = arg[i];
}

basis::basis()
{	rows = 4;
	columns = 4;
	values = new GLfloat[16];
	for(int i = 0; i < 16; i++)
		values[i] = 0;
}

basis::basis(const GLfloat * array)
{	rows = 4;
	columns = 4;
	values = new GLfloat[16];
	for(int i = 0; i < 16; i++)
		values[i] = array[i];
}

basis::basis(const matrix & arg)
{	int total = arg.getRows() * arg.getColumns();	// Total number of elements
	rows = arg.getRows();
	columns = arg.getColumns();
	values = new GLfloat[total];
	for(int i = 0; i < total; i++)
		values[i] = arg[i];
}

// The values in the matrix are assigned to the elements of the array.  Because no information is
// passed about the structure of the matrix, the assumption is that the row/column structure is
// the same as the matrix previously had.
void matrix::operator = (const GLfloat * right)
{	int total = rows * columns;
	for(int i = 0; i < total; i++)
		values[i] = right[i];
}

// The left matrix gets the same dimensions as the right one, and then the elements are copied in
// order.
void matrix::operator = (const matrix & right)
{	int total = right.rows * right.columns;
	if (rows * columns != total)
	{	if( rows > 0 && columns > 0)
			delete [] values;
		rows = right.rows;
		columns = right.columns;
		values = new GLfloat[total];
	}
	for(int i = 0; i < total; i++)
		values[i] = right.values[i];
}

// Multiplies the left (implicit) operand by the right, saving the answer in the left one.  This generally
// involves a change of dimension for the left matrix.
// The return value will be true if the multiplication was done.  If the matrices could not be multiplied
// (because the columns of the left did not equal the rows of the right) then it returns false and the
// implicit variable is unchanged.
bool matrix::operator *= (const matrix & right)
{	if(columns != right.rows)
		return false;
	
	matrix temp(rows, right.columns);
	for(int i = 0; i < rows; i++)					// For each row of the new matrix....
		for(int j = 0; j < right.columns; j++)		// For each column of the new matrix...
			for(int k = 0; k < columns; k++)		// We calculate the value of that entry by
										// doing the multiplications one by one.
				temp.values[i*temp.columns+j] += values[i*columns+k]*right.values[k*right.columns+j];
	*this = temp;
	return true;
}

// Adds the two matrices together, saving the answer in the left (implicit) one.  If they have different
// dimensions, this cannot be done and the function simply returns false.  If they can be added,
// however, true is returned.
bool matrix::operator += (const matrix & right)
{	if(columns != right.columns || rows != right.rows)
		return false;
	
	int total = rows * columns;				// Why bother with the illusion of a 2-dimensional
	for(int i = 0; i < total; i++)					// structure here?  Unlike multiplication, the structure
		values[i] += right.values[i];			// doesn't really effect the results...
	
	return true;
}

// Multiplies the left (implicit) operand by the right, returning the answer.  If the matrices could not be
// multiplied (because the columns of the left did not equal the rows of the right) then a matrix with
// 0 rows and 0 columns is returned.
matrix matrix::operator * (const matrix & right) const
{	if(columns != right.rows)
		return matrix(0,0);
	
	matrix res(rows, right.columns);			// res will hold the result of the multiplication
	for(int i = 0; i < rows; i++)					// For each row of res....
		for(int j = 0; j < right.columns; j++)		// For each column of res...
			for(int k = 0; k < columns; k++)		// We calculate the value of that entry by
										// doing the multiplications one by one.
				res.values[i*res.columns+j] += values[i*columns+k]*right.values[k*right.columns+j];

	return res;
}

// Adds the two matrices together, returning the answer.  If they have different dimensions and
// therefore cannot be added, then a matrix with 0 rows and 0 columns is returned.
matrix matrix::operator + (const matrix & right) const
{	if(columns != right.columns || rows != right.rows)
		return matrix(0,0);
	
	matrix res(rows, columns);				// res will hold the result of the addition
	int total = rows * columns;				// number of elements in the matrices
	for(int i = 0; i < total; i++)					// There is no point in bothering with the
		res.values[i] = values[i]+right.values[i];	// illusion of a 2-D structure here.  The
										// results are the same either way.
	return res;
}
// A mathematician would probably tell you that all matrices, even if they were a row or column vector,
// should be referenced with a 2-dimensional subscript.  But mathematicians are boring.  And they
// don't know enough about the limitations of C++ operators.  And at the moment, I've read a bit too
// much math and don't like mathematicians anyway.  So we're doing this with a single subscript.
// Yes, even 2-dimensional matrices use one subscript.
// Looking around the rest of this file, it should be pretty obvious that the class is using one array of
// size (rows*columns) and pretending it is a two-dimensional array of [rows][columns].  It gets away
// with this by converting the reference [a][b] into [a*columns+b].  So that's what you have to do to use
// this function.  Or you can use the index() class function whose sole purpose is to make that
// conversion for you.
// For the most part, though, I don't need to access elements of 2-dimensional matrices.  They're all
// either 1xn or mx1.  In those cases, I can pretend that it's a normal array.
GLfloat & matrix::operator [ ] (unsigned int index) const
{	if(index < rows*columns)			// If it's a valid element being requested...
		return values[index];			// ...then return it as a reference.
	return /*NULL;*/values[rows*columns-1];					// Otherwise, return NULL.  Let's see those smarty-pants
								// programmers try to access THAT!  HAHAHAHAHAHA!
}

// If you need a comment to tell you what this function does, you shouldn't be reading this anyway.
unsigned int matrix::getRows() const
{	return rows;	}

// This functions is almost as self-explanatory as getRows() was.  (Though this one does use
// bigger words.)
unsigned int matrix::getColumns() const
{	return columns;	}

// Changes the number of rows and columns in the matrix to those specified in the arguments.
// Elements from the previous matrix are copied over if they can be, while new spots are given
// the value 0.  (In other words, if a 3x10 matrix is turned into a 7x5, the leftmost 3x5 section of the
// first matrix is copied into the top 3x5 elements of the new one.  The bottom 4x5 become 0, and
// we have lost everything in the original matrix from columns 6 through 10.)
// Matrices with 0 dimensions are permitted.  The values pointer will be deallocated, of course, and
// the value of the other dimension will be preserved, in case the user expects to reference it later.
void matrix::setDimensions(unsigned int r, unsigned int c)
{	matrix temp(r, c);
	int minr, minc;					// minr holds the smaller of the row values, minc the
	if(r < rows)					// smaller column value;
		minr = rows;
	else	minr = r;
	if(c < columns)
		minc = columns;
	else minc = c;
	
	// Now that all the data is set up, copy all possible elements from original to new matrix
	for(int i = 0; i < minr; i++)
		for(int j = 0; j < minc; j++)
			temp.values[i*c+j] = this->values[i*rows+j];
}

// This function converts a row and column into a single index that refers to the corresponding
// location in the values array.  This is not necessary, but may be useful to the user.
// No error checking is done.
unsigned int matrix::index(unsigned int r, unsigned int c) const
{	return r*columns + c;
}