/* This file contains the specifications for a matrix class and some children of it.  The children have
 * specific dimensions, which makes this more convenient from the point of view of the curve classes.
 * Note that the subscript array must be one-dimensional.  If you are going to access a 2-d array with
 * this, you must multiply your row index by the number of columns and add that to your column index. 
 * (The same way that you would if you had a 1-d array masquerading as a 2-d one.)  Since I don't need
 * to worry about individual entries for this program, it's not a problem.  Just in case, though, I have
 * provided a function called index(), which takes the values of a the row and column and converts them
 * into a single index to use for the [ ] operator.  In other words, if you want to do M[a][b], and you don't
 * feel comfortable doing M[a*columns+b] (which you should do), you can do M[index(a, b)];
 * 
 * You can use the assignment operator, =, with either a matrix or an array on the right.  If you are
 * assigning a matrix to the value of another matrix, then the first one will take on the same dimensions
 * as the second.  If you are assigning it to the values in an array, however, the dimensions will be
 * unchanged.  Multiplication and addition work as expected; note that if you try an illegal operation
 * (such as adding matrices of different sizes) the returned value will be a 0X0 matrix.
 */

#ifndef NEVIN_MATRIX
#define NEVIN_MATRIX

#include <iostream>
#include <OpenGL/gl.h>
#include <OpenGL/glu.h>
#include <GLUT/glut.h> 
using namespace std;

//#include <GL/glut.h>

class matrix
{	public:
		// Constructors and Destructors:
		matrix();
		matrix(unsigned int, unsigned int);
		matrix(unsigned int, unsigned int, GLfloat *);
		matrix(const matrix&);
		
		~matrix();
		
		// Operators:
		void operator = (const GLfloat *);
		void operator = (const matrix&);
		bool operator *= (const matrix&);
		bool operator += (const matrix&);
		matrix operator * (const matrix&)const;
		matrix operator + (const matrix&) const;
		GLfloat & operator [ ] (unsigned int) const;
		
		// Accessor functions (is that what these are called?)
		unsigned int getRows() const;
		unsigned int getColumns() const;
		
		// Functions to change values of private data:
		void setDimensions(unsigned int, unsigned int);
		// (No function is needed for setting elements of the matrix.  Use the subscript
		// operator along with an assignment.)
		
		// Other cool things:  (I will eventually have stuff for finding the inverse, etc...)
		unsigned int index(unsigned int, unsigned int) const;
				
	protected:
		unsigned int rows, columns;				// Dimensions of matrix
		GLfloat * values;						// Data in matrix
};

// A coefficient is a 4x1 matrix (like M*G, giving coefficients to an equation of a curve)
class coefficient : public matrix
{	public:
		// Only needs to have constructors.  The parent class does everything else.
		coefficient();
		coefficient(const GLfloat *);
		coefficient(const matrix &);
};

// A basis is a 4x4 matrix (like we use for the basis matrix of a curve)
class basis : public matrix
{	public:
		// Again, this only needs its own constructors.
		basis();
		basis(const GLfloat *);
		basis(const matrix &);
};

#endif